#' gen_var2stat
#'
#' @description Function to generate variables from statistics
#'
#' @param code Character string with a maximum length of 15 characters. Code from a GENESIS, Zensus 2022 or regionalstatistik.de object. Only one code per iteration.
#' @param database Character string. Indicator if the GENESIS ('genesis'), Zensus 2022 ('zensus') or regionalstatistik.de ('regio') database is called. Default option is 'all'.
#' @param area Character string. Indicator from which area of the database the results are called. In general, 'all' is the appropriate solution. Default option is 'all'. Not used for 'statistics'.
#' @param detailed Boolean. Indicator if the function should return the detailed output of the iteration including all object-related information or only a shortened output including only code and object title. Default option is 'FALSE'.
#' @param sortcriterion Character string. Indicator if the output should be sorted by 'code' or 'content'. This is a parameter of the API call itself. The default is 'code'.
#' @param pagelength Integer. Maximum length of results or objects (e.g., number of tables). Defaults to 500. Maximum of the databases is 25,000 objects.
#' @param error.ignore Boolean. Indicator if the function should stop if an error occurs or no object for the request is found or if it should produce a token as response. Default option is 'FALSE'. Important note: This does not apply to errors generated by the API call itself (especially server or client errors with status codes 5xx or 4xx). In this case, the function call will error even if error.ignore is set to TRUE.
#' @param verbose Boolean. Indicator if the output of the function should include detailed messages and warnings. Default option is 'TRUE'. Set the parameter to 'FALSE' to suppress additional messages and warnings.
#' @param ... Additional parameters for the API call. These parameters are only affecting the call itself, no further processing. For more details see `vignette("additional_parameter")`.
#'
#' @return A list with all recalled elements from the API. Based on the 'detailed' parameter it contains more or less information, but always includes the code of the object, the title, and the type of the object. This is done to facilitate further processing with the data. Attributes are added to the data.frame describing the search configuration for the returned output.
#' @export
#'
#' @examples
#' \dontrun{
#' # Find the variables of the statistic with the code "12411"
#' # with a detailed output
#' object <- gen_var2stat(code = "12411", detailed = TRUE)
#' }
#'
gen_var2stat <- function(code = NULL,
                         database = c("all", "genesis", "zensus", "regio"),
                         area = c("all", "public", "user"),
                         detailed = FALSE,
                         sortcriterion = c("code", "content"),
                         pagelength = 500,
                         error.ignore = FALSE,
                         verbose = TRUE,
                         ...) {

  caller <- as.character(match.call()[1])

  check_function_input(code = code,
                       detailed = detailed,
                       error.ignore = error.ignore,
                       sortcriterion = sortcriterion,
                       pagelength = pagelength,
                       database = database,
                       caller = caller,
                       verbose = verbose)

  # database_vector will hold a vector of the specified databases to query
  database_vector <- test_database_function(input = database,
                                            error.input = error.ignore,
                                            text = verbose)

  sortcriterion <- match.arg(sortcriterion)

  area <- match.arg(area)

  area <- switch(area, all = "all", public = "\u00F6ffentlich", user = "benutzer")

  #-----------------------------------------------------------------------------

  # Processing #
  res <- lapply(database_vector, function(db){

    if (isTRUE(verbose)) {

      info <- paste("Started the processing of", db, "database.")

      message(info)

    }

    #---------------------------------------------------------------------------

    if (db == "genesis" | db == "regio") {

      results_raw <- gen_api(endpoint = "catalogue/variables2statistic",
                             database = db,
                             username = gen_auth_get(database = db)$username,
                             password = gen_auth_get(database = db)$password,
                             name = code,
                             area = area,
                             pagelength = pagelength,
                             ...)

    } else {

      results_raw <- gen_api(endpoint = "catalogue/variables2statistic",
                             database = db,
                             username = gen_auth_get(database = db)$username,
                             password = gen_auth_get(database = db)$password,
                             name = code,
                             pagelength = pagelength,
                             ...)

    }

    results_json <- test_if_json(results_raw)

    empty_object <- test_if_error(results_json, para = error.ignore, verbose = verbose)

    if (isTRUE(empty_object)) {

      list_of_variables <- "No 'variables' object found for your request."

    } else if (isFALSE(empty_object)) {

      list_of_variables <- results_json$Status$Content

    } else if (empty_object == "DONE") {

      if (isTRUE(detailed)) {

        list_of_variables <- binding_lapply(results_json$List,
                                            characteristics = c("Code",
                                                                "Content",
                                                                "Type",
                                                                "Values",
                                                                "Information"))

      } else {

        list_of_variables <- binding_lapply(results_json$List,
                                            characteristics = c("Code",
                                                                "Content"))

      }

      list_of_variables$Object_Type <- "Variable"

      list_of_variables <- tibble::as_tibble(list_of_variables)

    }

    #---------------------------------------------------------------------------

    # Summary #
    list_resp <- list("Variables" = list_of_variables)

    attr(list_resp, "Code") <- results_json$Parameter$name
    attr(list_resp, "Database") <- db
    attr(list_resp, "Language") <- results_json$Parameter$language
    attr(list_resp, "Pagelength") <- results_json$Parameter$pagelength
    attr(list_resp, "Copyright") <- results_json$Copyright

    return(list_resp)

  })

  #-----------------------------------------------------------------------------

  res <- check_results(res)

  return(res)

}

#-------------------------------------------------------------------------------

#' gen_val2var
#'
#' @description Function to extract the possible values from a variable. Values for continuous variables are not extractable, which is why the function returns a warning message in this case.
#'
#' @param code Character string with a maximum length of 15 characters. Code from a GENESIS, Zensus 2022 or regionalstatistik.de object. Only one code per iteration.
#' @param database Character string. Indicator if the GENESIS ('genesis'), Zensus 2022 ('zensus') or regionalstatistik.de ('regio') database is called. Default option is 'all'.
#' @param area Character string. Indicator from which area of the database the results are called. In general, 'all' is the appropriate solution. Default option is 'all'. Not used for 'statistics'.
#' @param sortcriterion Character string. Indicator if the output should be sorted by 'code' or 'content'. This is a parameter of the API call itself. The default is 'code'.
#' @param pagelength Integer. Maximum length of results or objects (e.g., number of tables). Defaults to 500. Maximum of the databases is 25,000 objects.
#' @param error.ignore Boolean. Indicator if the function should stop if an error occurs or no object for the request is found or if it should produce a token as response. Default option is 'TRUE'. Important note: This does not apply to errors generated by the API call itself (especially server or client errors with status codes 5xx or 4xx). In this case, the function call will error even if error.ignore is set to TRUE.
#' @param verbose Boolean. Indicator if the output of the function should include detailed messages and warnings. Default option is 'TRUE'. Set the parameter to 'FALSE' to suppress additional messages and warnings.
#' @param ... Additional parameters for the API call. These parameters are only affecting the call itself, no further processing. For more details see `vignette("additional_parameter")`.
#'
#' @return A list with all recalled elements from the API.  Always includes the code of the object, the title, and the type of the object. This is done to facilitate further processing with the data. Attributes are added to the data.frame describing the search configuration for the returned output.
#' @export
#'
#' @examples
#' \dontrun{
#' # Find the values of the variable "DLAND"
#' object <- gen_val2var(code = "DLAND")
#' }
#'
gen_val2var <- function(code = NULL,
                        database = c("all", "genesis", "zensus", "regio"),
                        area = c("all", "public", "user"),
                        sortcriterion = c("code", "content"),
                        pagelength = 500,
                        error.ignore = TRUE,
                        verbose = TRUE,
                        ...) {

  caller <- as.character(match.call()[1])

  check_function_input(code = code,
                       error.ignore = error.ignore,
                       sortcriterion = sortcriterion,
                       pagelength = pagelength,
                       database = database,
                       caller = caller,
                       verbose = verbose)

  # database_vector will hold a vector of the specified databases to query
  database_vector <- test_database_function(input = database,
                                            error.input = error.ignore,
                                            text = verbose)

  sortcriterion <- match.arg(sortcriterion)

  area <- match.arg(area)

  area <- switch(area, all = "all", public = "\u00F6ffentlich", user = "benutzer")

  embedding <- deparse(sys.calls())

  #-----------------------------------------------------------------------------

  res <- lapply(database_vector, function(db){

    if (isTRUE(verbose)) {

      info <- paste("Started the processing of", db, "database.")

      message(info)

    }

    if (db == "genesis" | db == "regio") {

      results_raw <- gen_api(endpoint = "catalogue/values2variable",
                             database = db,
                             username = gen_auth_get(database = db)$username,
                             password = gen_auth_get(database = db)$password,
                             name = code,
                             area = area,
                             pagelength = pagelength,
                             ...)

    } else {

      results_raw <- gen_api(endpoint = "catalogue/values2variable",
                             database = db,
                             username = gen_auth_get(database = db)$username,
                             password = gen_auth_get(database = db)$password,
                             name = code,
                             pagelength = pagelength,
                             ...)

    }

    results_json <- test_if_json(results_raw)

    if (isFALSE(grepl("pairlist\\(gen_val2var", embedding))) {

      empty_object <- test_if_error_variables(results_json, para = error.ignore)

    } else {

      empty_object <- test_if_error(results_json, para = error.ignore, verbose = verbose)

    }

    if (isTRUE(empty_object)) {

      list_of_variables <- "No 'values' object found for your request."

    } else if (isFALSE(empty_object)) {

      list_of_variables <- results_json$Status$Content

    } else if (empty_object == "DONE") {

      list_of_variables <- binding_lapply(results_json$List,
                                          characteristics = c("Code",
                                                              "Content",
                                                              "Variables",
                                                              "Information"))

      list_of_variables$Object_Type <- "Value"

      list_of_variables <- tibble::as_tibble(list_of_variables)

    }

    list_resp <- list("Values" = list_of_variables)

    attr(list_resp, "Name") <- results_json$Parameter$name
    attr(list_resp, "Database") <- db
    attr(list_resp, "Language") <- results_json$Parameter$language
    attr(list_resp, "Pagelength") <- results_json$Parameter$pagelength
    attr(list_resp, "Copyright") <- results_json$Copyright

    names(list_resp) <- paste("Values of", results_json$Parameter$name)

    return(list_resp)

  })

  #-----------------------------------------------------------------------------

  res <- check_results(res)

  return(res)

}

#-------------------------------------------------------------------------------

#' gen_val2var2stat
#'
#' @description Get values from variables from a statistic. Values for continuous variables cannot be extracted, so the function returns a warning message.
#'
#' @param code Character string with a maximum length of 15 characters. Code from a GENESIS, Zensus 2022 or regionalstatistik.de object. Only one code per iteration.
#' @param database Character string. Indicator if the GENESIS ('genesis'), Zensus 2022 ('zensus') or regionalstatistik.de ('regio') database is called. Default option is 'all'.
#' @param area Character string. Indicator from which area of the database the results are called. In general, 'all' is the appropriate solution. Default option is 'all'. Not used for 'statistics'.
#' @param detailed Boolean. Indicator if the function should return the detailed output of the iteration including all object-related information or only a shortened output including only code and object title. Default option is 'FALSE'.
#' @param sortcriterion Character string. Indicator if the output should be sorted by 'code' or 'content'. This is a parameter of the API call itself. The default is 'code'.
#' @param pagelength Integer. Maximum length of results or objects (e.g., number of tables). Defaults to 500. Maximum of the databases is 25,000 objects.
#' @param error.ignore.var Boolean. Indicator for variables if the function should stop if an error occurs or no object for the request is found or if it should produce a token as response. Default option is 'FALSE'. Important note: This does not apply to errors generated by the API call itself (especially server or client errors with status codes 5xx or 4xx). In this case, the function call will error even if error.ignore is set to TRUE.
#' @param error.ignore.val Boolean. Indicator for values if the function should stop if an error occurs or no object for the request is found or if it should produce a token as response. Default option is 'TRUE', this prevents the function to stop even if a variable has no further explanation (often the case for numerical variables). Important note: This does not apply to errors generated by the API call itself (especially server or client errors with status codes 5xx or 4xx). In this case, the function call will error even if error.ignore is set to TRUE.
#' @param verbose Boolean. Indicator if the output of the function should include detailed messages and warnings. Default option is 'TRUE'. Set the parameter to 'FALSE' to suppress additional messages and warnings.
#' @param ... Additional parameters for the API call. These parameters are only affecting the call itself, no further processing. For more details see `vignette("additional_parameter")`.
#'
#' @return A list with all recalled elements from the API. Based on the 'detailed' parameter it contains more or less information, but always includes the code of the object, the title, and the type of the object. This is done to facilitate further processing with the data. Attributes are added to the data.frame describing the search configuration for the returned output.
#' @export
#'
#' @examples
#' \dontrun{
#' # Find the values of variables in a specific statistic with
#' # the code "21111" and a detailed description of the variables
#' object <- gen_val2var2stat(code = "21111", detailed = TRUE)
#' }
#'
gen_val2var2stat <- function(code = NULL,
                             database = c("all", "genesis", "zensus", "regio"),
                             area = c("all", "public", "user"),
                             detailed = FALSE,
                             sortcriterion = c("code", "content"),
                             pagelength = 500,
                             error.ignore.var = FALSE,
                             error.ignore.val = TRUE,
                             verbose = TRUE,
                             ...) {

  caller <- as.character(match.call()[1])

  check_function_input(code = code,
                       error.ignore = error.ignore.var,
                       sortcriterion = sortcriterion,
                       pagelength = pagelength,
                       database = database,
                       caller = caller,
                       verbose = verbose)

  # database_vector will hold a vector of the specified databases to query
  database_vector <- test_database_function(input = database,
                                            error.input = error.ignore.var,
                                            text = verbose)

  sortcriterion <- match.arg(sortcriterion)

  #-----------------------------------------------------------------------------

  res <- lapply(database_vector, function(db){

    variables <- suppressMessages(suppressWarnings(gen_var2stat(code = code,
                                                                database = database_vector,
                                                                area = area,
                                                                detailed = detailed,
                                                                sortcriterion = sortcriterion,
                                                                error.ignore = error.ignore.var,
                                                                verbose = verbose,
                                                                ...)))

    if (length(dim(variables$Variables)) != 2) {

      if (variables$Variables == "No 'variables' object found for your request.") {

        list_resp <- variables

      }

    } else {

      list_values <- list()

      lapply(variables$Variables$Code, function(x) {

        zwisch <- suppressMessages(suppressWarnings(gen_val2var(code = x,
                                                                database = database_vector,
                                                                area = area,
                                                                sortcriterion = sortcriterion,
                                                                error.ignore = error.ignore.val,
                                                                verbose = verbose)))

        list_values <<- append(list_values, zwisch)

      })

      list_resp <- list(variables, list_values)

    }

    return(list_resp)

  })

  #-----------------------------------------------------------------------------

  res <- check_results(res)

  return(res)

}

#-------------------------------------------------------------------------------

#' gen_search_vars
#'
#' @description Function to search for specific variables
#'
#' @param code Character string with a maximum length of 6 characters. Code from a GENESIS, Zensus 2022 or regionalstatistik.de object. Only one code per iteration.
#' @param database Character string. Indicator if the GENESIS ('genesis'), Zensus 2022 ('zensus') or regionalstatistik.de ('regio') database is called. Default option is 'all'.
#' @param area Character string. Indicator from which area of the database the results are called. In general, 'all' is the appropriate solution. Default option is 'all'. Not used for 'statistics'.
#' @param sortcriterion Character string. Indicator if the output should be sorted by 'code' or 'content'. This is a parameter of the API call itself. The default is 'code'.
#' @param pagelength Integer. Maximum length of results or objects (e.g., number of tables). Defaults to 500. Maximum of the databases is 25,000 objects.
#' @param error.ignore Boolean. Indicator if the function should stop if an error occurs or no object for the request is found or if it should produce a token as response. Default option is 'FALSE'. Important note: This does not apply to errors generated by the API call itself (especially server or client errors with status codes 5xx or 4xx). In this case, the function call will error even if error.ignore is set to TRUE.
#' @param verbose Boolean. Indicator if the output of the function should include detailed messages and warnings. Default option is 'TRUE'. Set the parameter to 'FALSE' to suppress additional messages and warnings.
#' @param ... Additional parameters for the API call. These parameters are only affecting the call itself, no further processing. For more details see `vignette("additional_parameter")`.
#'
#' @return A list with all recalled elements from the API. Always includes the code of the object, the title, and the type of the object. This is done to facilitate further processing with the data. Attributes are added to the data.frame describing the search configuration for the returned output.
#' @export
#'
#' @examples
#' \dontrun{
#' # Find a specific variable "GES" in Genesis
#' object <- gen_search_vars("GES")
#' }
#'
gen_search_vars <- function(code = NULL,
                            database = c("all", "genesis", "zensus", "regio"),
                            area = c("all", "public", "user"),
                            sortcriterion = c("code", "content"),
                            pagelength = 500,
                            error.ignore = FALSE,
                            verbose = TRUE,
                            ...) {

  caller <- as.character(match.call()[1])

  check_function_input(code = code,
                       error.ignore = error.ignore,
                       sortcriterion = sortcriterion,
                       pagelength = pagelength,
                       database = database,
                       caller = caller,
                       verbose = verbose)

  # database_vector will hold a vector of the specified databases to query
  database_vector <- test_database_function(input = database,
                                            error.input = error.ignore,
                                            text = verbose)

  sortcriterion <- match.arg(sortcriterion)

  area <- match.arg(area)

  area <- switch(area, all = "all", public = "\u00F6ffentlich", user = "benutzer")

  #-----------------------------------------------------------------------------

  res <- lapply(database_vector, function(db){

    if (isTRUE(verbose)) {

      info <- paste("Started the processing of", db, "database.")

      message(info)

    }

    #---------------------------------------------------------------------------

    if (db == "genesis" | db == "regio") {

      results_raw <- gen_api(endpoint = "catalogue/variables",
                             database = db,
                             username = gen_auth_get(database = db)$username,
                             password = gen_auth_get(database = db)$password,
                             name = code,
                             sortcriterion = sortcriterion,
                             area = area,
                             pagelength = pagelength,
                             ...)

    } else {

      results_raw <- gen_api(endpoint = "catalogue/variables",
                             database = db,
                             username = gen_auth_get(database = db)$username,
                             password = gen_auth_get(database = db)$password,
                             sortcriterion = sortcriterion,
                             name = code,
                             pagelength = pagelength,
                             ...)

    }

    results_json <- test_if_json(results_raw)

    empty_object <- test_if_error(results_json, para = error.ignore)

    if (isTRUE(empty_object)) {

      list_of_variables <- "No 'variables' object found for your request."

    } else if (isFALSE(empty_object)) {

      list_of_variables <- results_json$Status$Content

    } else if (empty_object == "DONE") {

      list_of_variables <- binding_lapply(results_json$List,
                                          characteristics = c("Code",
                                                              "Content",
                                                              "Type",
                                                              "Information"))

      list_of_variables$Object_Type <- "Variable"

      list_of_variables <- tibble::as_tibble(list_of_variables)

    }

    list_resp <- list("Variables" = list_of_variables)

    attr(list_resp, "Code") <- results_json$Parameter$selection
    attr(list_resp, "Database") <- database[1]
    attr(list_resp, "Language") <- results_json$Parameter$language
    attr(list_resp, "Pagelength") <- results_json$Parameter$pagelength
    attr(list_resp, "Copyright") <- results_json$Copyright

    return(list_resp)

  })

  #-----------------------------------------------------------------------------

  res <- check_results(res)

  return(res)

}
